﻿using System;
using SDX.Compiler;
using Mono.Cecil;
using Mono.Cecil.Cil;
using System.Linq;
using System.Collections.Generic;

public class SilentAttackFix : IPatcherMod
{
    public bool Link(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        InsertGameStageDefinitionRead(gameModule, modModule);
        InsertEntityAliveConditionalTriggerSleeperWakeUp(gameModule, modModule);
        InsertEntityAliveAwake(gameModule, modModule);
        InsertEntityAliveSetAttackTarget(gameModule, modModule);
        InsertEntityAliveOnUpdateEntity(gameModule, modModule);
        InsertEntityAliveAttack(gameModule, modModule);
        return true;
    }
    public void InsertEntityAliveAttack(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "EntityAlive");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Attack");

        Instruction insertPosition = null;
        Instruction jumpToPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Boolean EntityAlive::VT")
            {
                insertPosition = inst.Previous;
                jumpToPosition = inst.Next.Next;
            }
        }
        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }
        TypeDefinition silentAttackFixMethod = modModule.Types.First(d => d.Name == "SilentAttackFixMethod");
        MethodDefinition isSilentAttackFix = silentAttackFixMethod.Methods.First(d => d.Name == "IsSilentAttackFix");

        MethodReference isSilentAttackFixRef = gameModule.Import(isSilentAttackFix);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        Instruction newJumpTo = Instruction.Create(OpCodes.Call, isSilentAttackFixRef);
        proc1.InsertBefore(insertPosition, newJumpTo);
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Brtrue_S, jumpToPosition));

        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }

    }
    public void InsertEntityAliveOnUpdateEntity(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "EntityAlive");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "OnUpdateEntity");

        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Int32 EntityAlive::attackingTime")
            {
                if (inst.Next.Next.OpCode == OpCodes.Bgt)
                {
                    insertPosition = inst.Previous;
                }
            }
        }
        TypeDefinition silentAttackFixMethod = modModule.Types.First(d => d.Name == "SilentAttackFixMethod");
        MethodDefinition attackAlert = silentAttackFixMethod.Methods.First(d => d.Name == "AttackAlert");
        MethodReference attackAlertRef = gameModule.Import(attackAlert);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, attackAlertRef));

    }
    public void InsertEntityAliveSetAttackTarget(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "EntityAlive");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "SetAttackTarget");

        Instruction insertPosition = null;
        Instruction jumpToPosition = null;
        FieldDefinition ulField = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {

            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Int32 UnityEngine.Random::Range(System.Int32,System.Int32)")
            {
                insertPosition = inst.Previous.Previous.Previous;

            }
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Int32 EntityAlive::UL")
            {
                ulField = (FieldDefinition)inst.Operand;
                jumpToPosition = inst.Next;
            }
        }
        TypeDefinition silentAttackFixMethod = modModule.Types.First(d => d.Name == "SilentAttackFixMethod");
        MethodDefinition isSilentAttackFix = silentAttackFixMethod.Methods.First(d => d.Name == "IsSilentAttackFix");

        MethodReference isSilentAttackFixRef = gameModule.Import(isSilentAttackFix);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(jumpToPosition, Instruction.Create(OpCodes.Br_S, jumpToPosition));
        Instruction jumpToPos2 = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(jumpToPosition, jumpToPos2);
        proc1.InsertBefore(jumpToPosition, Instruction.Create(OpCodes.Ldc_I4_0));
        proc1.InsertBefore(jumpToPosition, Instruction.Create(OpCodes.Stfld, ulField));


        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, isSilentAttackFixRef));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Brtrue_S, jumpToPos2));


    }
    public void InsertEntityAliveAwake(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "EntityAlive");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Awake");

        Instruction insertPosition = null;
        Instruction jumpToPosition = null;
        FieldDefinition ulField = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {

            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single Entity::stepHeight")
            {
                insertPosition = inst.Next;

            }
            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Int32 EntityAlive::UL")
            {
                ulField = (FieldDefinition)inst.Operand;
                jumpToPosition = inst.Next;
            }
        }
        TypeDefinition silentAttackFixMethod = modModule.Types.First(d => d.Name == "SilentAttackFixMethod");
        MethodDefinition isSilentAttackFix = silentAttackFixMethod.Methods.First(d => d.Name == "IsSilentAttackFix");

        MethodReference isSilentAttackFixRef = gameModule.Import(isSilentAttackFix);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(jumpToPosition, Instruction.Create(OpCodes.Br_S, jumpToPosition));
        Instruction jumpToPos2 = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(jumpToPosition, jumpToPos2);
        proc1.InsertBefore(jumpToPosition, Instruction.Create(OpCodes.Ldc_I4_0));
        proc1.InsertBefore(jumpToPosition, Instruction.Create(OpCodes.Stfld, ulField));


        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, isSilentAttackFixRef));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Brtrue_S, jumpToPos2));

    }
    public void InsertEntityAliveConditionalTriggerSleeperWakeUp(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "EntityAlive");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "ConditionalTriggerSleeperWakeUp");
        FieldDefinition sleeperSupressLivingSoundsField = classDef.Fields.First(d => d.Name == "SleeperSupressLivingSounds");

        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {

            if (inst.OpCode == OpCodes.Stfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Boolean EntityAlive::IsSleeping")
            {
                insertPosition = inst.Next;

            }
        }

        TypeDefinition silentAttackFixMethod = modModule.Types.First(d => d.Name == "SilentAttackFixMethod");
        MethodDefinition isSilentAttackFix = silentAttackFixMethod.Methods.First(d => d.Name == "IsSilentAttackFix");

        MethodReference isSilentAttackFixRef = gameModule.Import(isSilentAttackFix);

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, isSilentAttackFixRef));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Brfalse_S, insertPosition));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Ldc_I4_0));
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Stfld, sleeperSupressLivingSoundsField));


    }
    public void InsertGameStageDefinitionRead(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "GameStagesFromXml");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Load");

        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {

            if (inst.OpCode == OpCodes.Stsfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single GameStageDefinition::DaysAliveDifficultyBonus")
            {
                insertPosition = inst.Next;

            }
        }
        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Brfalse_S && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[4]);
        proc1.InsertBefore(insertPosition, newJumpTo);//V_4 is XmlElement

        TypeDefinition silentAttackFixMethod = modModule.Types.First(d => d.Name == "SilentAttackFixMethod");
        MethodDefinition setAddAttribute = silentAttackFixMethod.Methods.First(d => d.Name == "SetAddAttribute");

        MethodReference setAddAttributeRef = gameModule.Import(setAddAttribute);

        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, setAddAttributeRef));

        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }
    }
    public bool Patch(ModuleDefinition module)
    {
        Console.WriteLine("==SilentAttackFix Patcher===");
        SetEntityAlivePublic(module);
        return true;
    }

    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        classDef.Fields.Add(fd);

    }
    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType, object constantvalue)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        fd.Constant = constantvalue;
        classDef.Fields.Add(fd);

    }
    private void SetEntityAlivePublic(ModuleDefinition module)
    {

        TypeDefinition classDef = module.Types.First(d => d.Name == "EntityAlive");
        SetFieldToPublic(classDef, "attackingTime");
        SetFieldToPublic(classDef, "UL");
        SetFieldToPublic(classDef, "IL");
        
    }
    private void SetFieldToPublic(TypeDefinition classType, string fieldName)
    {
        SetFieldToPublic(classType.Fields.First(d => d.Name == fieldName));
    }
    private void SetFieldToPublic(FieldDefinition field)
    {
        field.IsFamily = false;
        field.IsPrivate = false;
        field.IsPublic = true;

    }

}
